#
# Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
import time
import subprocess
import fcntl
import json
import re
import yaml

import schedule
import eventlet
import eventlet.pools

from netq_lib.common import utils, configuration
from netq_lib.common.enums import ConfKw, DefaultPaths, NetqLogGroup, spiceAmberGroups, spiceAmberSeqNo
from netq_lib.common.utils import send_list_data, get_chip_type_v4

from netq_lib.orm.redisdb.models import (
    SpiceAmberGlobalV2, SpiceAmberSysLong, SpiceAmberSysShort, SpiceAmberSysCritical,
    SpiceAmberModShort, SpiceAmberModLong, SpiceAmberModCritical, SpiceAmberGlobalSysCritical
)
from netq_agent.handlers.port_handler import PortHandler
from netq_agent.handlers.sdk_utils import SdkUtils, SX_CHIP_TYPE_SPECTRUM4

_PROC_NAME = 'netq-agent'

class SpiceDefaultConfiguration(object):
    """
    class contains default whitelist of amBer directories
    """

    def __init__(self):
        self.default_config = {
            'global': {
                'parallelism': '1',
                'period': '300',
                'timeout': '1',
                'read-fails': '3',
                'whitelist': [
                    {'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/MDIR'},
                    {'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/MGIR'},
                    {'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/MSGI'},
                    {'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/SPAD'},
                ]
            },
            'local': [
                {'whitelist': {
                    'period': '120',
                    'read-fails': '3',
                    'timeout': '1',
                    'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/local_port.1/*'
                    }
                }
            ]
        }


class _SpiceConfig(object):
    """Spice Config class.
    """
    _DEF_TIMEOUT = 5  # End the cmd if not finished within this time
    _RANDOM_START = 20
    _RANDOM_STOP = 60
    _RANDOM_STEP = 10
    _PATH_RESET_PERIOD = 24 * 60 * 60

    def __init__(self, xpath, period, read_fails, logger, timeout=None, is_global=False):
        self._logger = logger
        self.xpath = xpath
        self.period = period
        self.read_fail_allowed = read_fails
        self.isGlobal = is_global
        self.timeout = timeout if (timeout is not None) else self._DEF_TIMEOUT
        self.path_set = set()
        self.skipped_path = dict()
        self._expand_xpath()
        self._path_reset_count = 0

    def _get_list_of_files(self, dir_name):
        """
        create a list of file and subdirectories
        names in the given directory
        """

        if os.path.isfile(dir_name):
            return [dir_name]
        list_of_files = os.listdir(dir_name)
        all_files = list()
        # Iterate over all the entries
        for entry in list_of_files:
            # Create full path
            full_path = os.path.join(dir_name, entry)
            # If entry is a directory then get the list of files in this directory
            if os.path.isdir(full_path):
                all_files = all_files + self._get_list_of_files(full_path)
            else:
                all_files.append(full_path)
        return all_files

    def _expand_xpath(self):
        if "*" in self.xpath:
            self.xpath = self.xpath.replace("*", "")
        if not os.path.exists(self.xpath):
            return
        path = self._get_list_of_files(self.xpath)
        self.path_set = set(path)
        self._logger.debug("spice: total {} path in xpath: {}".format(len(self.path_set), self.xpath))

    def mark_path_skipped(self, path):
        count = self.skipped_path.setdefault(path, 0)
        self.skipped_path[path] = count+1

    def is_path_skipped(self, path):
        if path in self.skipped_path.keys():
            return self.skipped_path[path] > self.read_fail_allowed
        return False

    def reset_skipped_paths(self):
        if self._path_reset_count >= self._PATH_RESET_PERIOD:
            self._path_reset_count = 0
            self._logger.debug("spice: skipped paths: {} before reset in xpath: {}".format(len(self.skipped_path), self.xpath))
            self.skipped_path.clear()
        else:
            self._path_reset_count = self._path_reset_count + self.period

    def force_reset_skipped_paths(self):
        self.skipped_path.clear()

    def exclude_local(self, local_path_obj):
        self.path_set = self.path_set - local_path_obj.path_set


class SpiceCollectionAgent(object):
    """
    Main class for Spice collection Framework which
    collects and stream data periodically
    """
    _BER_ROOT = '/sys/kernel/debug/nvswitch/ber/'
    _CHECK_PORTMAP_INTERVAL = 290
    _SDK_RESTART_DELAY = 30
    _DEFAULT_BATCH_COUNT = 100

    spice_groups_max = spiceAmberGroups.SPICE_AMBER_GROUP_MAX.value
    spice_dbg_last_modified_time = None
    is_spice_dbg_dump_enabled = [0 for _ in range(spice_groups_max)]
    # Loop index in sequence obj tracks number of spice_ts, resets if spice subprocess restarts
    seqno_loop_idx = spiceAmberSeqNo.SPICE_AMBER_SEQNO_LOOP_IDX.value
    seqno_batch_idx = spiceAmberSeqNo.SPICE_AMBER_SEQNO_BATCH_IDX.value
    seqno_path_idx = spiceAmberSeqNo.SPICE_AMBER_SEQNO_PATH_IDX.value
    seqno_max_idx = spiceAmberSeqNo.SPICE_AMBER_SEQNO_MAX_IDX.value

    def __init__(self, logger, platform):
        self._logger = logger
        self._platform = platform
        self._is_spice_loaded = False
        self._config_dict = {}
        self._template_config_dict = {}
        self._pool = eventlet.GreenPool()
        self._timers = {}
        self._spice_path_objs = []
        self._global_paths = []
        self._local_paths = []
        self.sx_sdk_pid = -1
        self._default_config_obj = SpiceDefaultConfiguration()
        self.port_handler = PortHandler.get_instance(self._logger, None)
        self.sdk_utils = SdkUtils(self._logger)
        self.mapped_local_ports = []
        self.__logger_coalescer = utils.LogCoalescer("netq-agent", conf_dict=configuration.get_config(_PROC_NAME))
        self.__path_to_model_dict = {}

    def __load_configuration(self):
        chip_type = get_chip_type_v4(self._platform)
        config_file = DefaultPaths.SPICE_CONFIG_FILE_SPC4 if chip_type == SX_CHIP_TYPE_SPECTRUM4 else DefaultPaths.SPICE_CONFIG_FILE_DEFAULT
        template_config_file = DefaultPaths.SPICE_TEMPLATE_CONFIG_FILE_SPC4 if chip_type == SX_CHIP_TYPE_SPECTRUM4 else DefaultPaths.SPICE_TEMPLATE_CONFIG_FILE_DEFAULT

        if os.path.exists(config_file):
            self._logger.debug(f"spice: Loading configuration from {config_file}")
            with open(config_file, 'r') as f:
                self._config_dict = yaml.safe_load(f)
        else:
            self._logger.info("spice: Missing spice configuration")
            return False

        return True

    def __get_index_for_model(self, model):
        model_index = None
        if model == 'SpiceAmberSysCritical':
            model_index = spiceAmberGroups.SPICE_AMBER_SYS_CRITICAL.value
        elif model == 'SpiceAmberModCritical':
            model_index = spiceAmberGroups.SPICE_AMBER_MOD_CRITICAL.value
        elif model == 'SpiceAmberSysLong':
            model_index = spiceAmberGroups.SPICE_AMBER_SYS_LONG.value
        elif model == 'SpiceAmberSysShort':
            model_index = spiceAmberGroups.SPICE_AMBER_SYS_SHORT.value
        elif model == 'SpiceAmberModLong':
            model_index = spiceAmberGroups.SPICE_AMBER_MOD_LONG.value
        elif model == 'SpiceAmberModShort':
            model_index = spiceAmberGroups.SPICE_AMBER_MOD_SHORT.value
        elif model == 'SpiceAmberGlobalV2':
            model_index = spiceAmberGroups.SPICE_AMBER_GLOBALV2.value
        elif model == 'SpiceAmberGlobalSysCritical':
            model_index = spiceAmberGroups.SPICE_AMBER_GLOBAL_SYS_CRITICAL.value
        self._logger.debug(f"spice: index for model - {model} {model_index}")
        return model_index

    def __get_class_for_model(self, model):
        model_class = None
        if model == 'SpiceAmberSysCritical':
            model_class = SpiceAmberSysCritical()
        elif model == 'SpiceAmberModCritical':
            model_class = SpiceAmberModCritical()
        elif model == 'SpiceAmberSysLong':
            model_class = SpiceAmberSysLong()
        elif model == 'SpiceAmberSysShort':
            model_class = SpiceAmberSysShort()
        elif model == 'SpiceAmberModLong':
            model_class = SpiceAmberModLong()
        elif model == 'SpiceAmberModShort':
            model_class = SpiceAmberModShort()
        elif model == 'SpiceAmberGlobalV2':
            model_class = SpiceAmberGlobalV2()
        elif model == 'SpiceAmberGlobalSysCritical':
            model_class = SpiceAmberGlobalSysCritical()
        return model_class

    def __get_model_for_path(self, path):
        if path in self.__path_to_model_dict:
            (model, index) = self.__path_to_model_dict[path]
            self._logger.debug(f"spice: model for path: {path} {model} {index}")
            return model, index
        self._logger.debug(f"spice: path {path} not in dict")
        return None, None

    def __set_model_for_path(self, path, model):
        if path not in self.__path_to_model_dict:
            index = self.__get_index_for_model(model)
            self.__path_to_model_dict[path] = (model, index)
            self._logger.debug(f"spice: set model for path: {path} model: {model} index: {index}")
            return model, index
        return None, None

    def __load_spice_module(self):
        if os.path.exists(self._BER_ROOT):
            return True
        load_cmd = "/sbin/modprobe sx_spice_dump"
        try:
            utils.subprocess_wrapper(load_cmd.split(' '))
            time.sleep(1)
            if os.path.exists(self._BER_ROOT):
                return True
        except subprocess.CalledProcessError as ex:
            self._logger.warn("spice: unable to load spice kernel module: {}".format(ex.returncode))

        return False

    @staticmethod
    def __is_path_port_level(path):
        if 'local_port' in path:
            return True
        return False

    def __extract_port_from_path(self, path, is_fetch_name):
        port_num = 0
        port_name = 'n/a'

        if not self.__is_path_port_level(path):
            return port_num, port_name
        path_parents = path.split("/")
        try:
            for p in path_parents:
                if 'local_port' in p:
                    port_num = int(p.split(".")[1])
                    if is_fetch_name:
                        lid_port_num = port_num + 0x10000
                        port_name = self.port_handler.get_port_name(lid_port_num)
        except Exception:
            port_num = 0

        return port_num, port_name

    def _resize_green_pool(self):
        parallel_runs = 1
        self._pool.resize(parallel_runs)

    def expand_range(self, path_regex):
        if "*" in path_regex:
            path_regex = path_regex.replace("*", "")

        matches = re.findall(r'\[(\d+)-(\d+)\]', path_regex)
        expanded_xpaths = [path_regex]
        if matches:
            for match in matches:
                start = int(match[0])
                end = int(match[1])
                new_expanded_xpaths = []
                # Expand the current range in each existing xpath
                for entry in expanded_xpaths:
                    new_expanded_xpaths.extend([entry.replace(f"[{start}-{end}]", f"{i}") for i in range(start, end+1)])

                expanded_xpaths = new_expanded_xpaths
        return expanded_xpaths

    def _append_paths(self, expanded_xpaths, period, timeout, read_fails, model, is_global_path):
        for expanded_xpath in expanded_xpaths:
            if not os.path.exists(expanded_xpath):
                self._logger.debug("spice: {} not exist".format(expanded_xpath))
                continue
            spice_path_obj = _SpiceConfig(expanded_xpath, period, read_fails, self._logger, timeout, is_global=is_global_path)

            if is_global_path:
                self._global_paths.append(spice_path_obj)
            else:
                self._local_paths.append(spice_path_obj)

            self.__set_model_for_path(expanded_xpath, model)

            self._spice_path_objs.append(spice_path_obj)

    def _setup_path_whitelist(self):
        min_period = 3600
        for section, section_data in self._config_dict.items():
            # Process local group settings
            if section == 'spice':
                spice_section = section_data
                whitelists = spice_section.get("whitelists", [])
                for whitelist in whitelists:
                    period = int(whitelist.get('period'))
                    if period < min_period:
                        min_period = period
                    read_fails = int(whitelist.get('read-fails'))
                    timeout = int(whitelist.get('timeout'))
                    scope = str(whitelist.get('scope'))
                    if "global" in scope:
                        is_global = True
                    else:
                        is_global = False

                    models = whitelist.get("groups", [])
                    for model in models:
                        xpaths = model.get("xpaths", [])
                        for xpath in xpaths:
                            xpath_value = xpath.get('xpath')
                            expanded_xpaths = self.expand_range(xpath_value)
                            self._append_paths(expanded_xpaths, period, timeout, read_fails, model.get('group'), is_global)

        self._logger.debug("spice: Completed setup_path_whitelist: {}".format(min_period))
        return min_period

    def __update_path_timer(self, path, epoch):
        epoch_new = int(epoch)
        self._timers[path] = epoch_new + path.period

    def __config_path_timers(self):
        self._timers = {}
        epoch = int(time.time())
        for path in self._spice_path_objs:
            self._timers[path] = epoch

    def __get_path_data(self, path):
        file_dict = {}
        try:
            with open(path, "r") as file:
                for line in file:
                    line = line.strip()
                    if 'N/A' in line:
                        file_dict['NA'] = True
                        continue
                    if len(line) == 0:  # pylint: disable=len-as-condition
                        continue
                    line_list = line.split(":")
                    if len(line_list) < 2:
                        continue
                    file_dict[line_list[0].strip()] = line_list[-1].strip()
        except Exception as e:
            pass
        return file_dict

    def __process_spice_path(self, path_obj, path, epoch):
        try:
            if path_obj.is_path_skipped(path):
                return None, None
            with eventlet.timeout.Timeout(path_obj.timeout, False):
                value = self.__get_path_data(path)
                eventlet.sleep(0)
            if not value:
                self.__logger_coalescer.warn("SPICE: unable to read path ", path, group=NetqLogGroup.SPICE_PATH_WARNING)
                path_obj.mark_path_skipped(path)
                return None, None
            if 'NA' in value.keys() and value['NA']:
                path_obj.mark_path_skipped(path)
                return None, None

            model = None

            model_name, index = self.__get_model_for_path(path)

            if model_name is None:
                self._logger.error("spice: Cannot find model for path {0}".format(path))
                return None, None

            model = self.__get_class_for_model(model_name)

            if not model:
                self._logger.error("spice: Cannot find model class for path {0}".format(path))
                return None, None

            if self.__is_path_port_level(path):
                port_num, port_name = self.__extract_port_from_path(path, True)
                model.port_num = port_num
                model.port_name = port_name

            model_seqno = self.model_seqno_list[index]
            model.spice_path = path
            model.spice_value = value
            model.spice_ts = epoch
            model.spice_seqno_path = model_seqno[self.seqno_path_idx]
            model.spice_seqno_batch = model_seqno[self.seqno_batch_idx]
            model.spice_seqno_loop = model_seqno[self.seqno_loop_idx]
            model_seqno[self.seqno_path_idx] += 1
            return model, index

        except Exception as e:
            self._logger.debug("spice: caught Exception {0} for path {1}".format(e, path))
            pass

    def _log_current_monitored_paths(self):
        mode = 'w'
        dir_path = ConfKw.DEFAULT_NETQ_AGENT_STATUS_DIR
        file_name = "pruned_spice_paths"
        file_path = dir_path + file_name

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        with open(file_path, mode) as f:
            fcntl.flock(f, fcntl.LOCK_EX)
            for spice_path in self._spice_path_objs:
                f.write("{} \n".format(spice_path.xpath))
            fcntl.flock(f, fcntl.LOCK_UN)

    def _update_local_path_list(self):
        """
        Get the front panel to sdk local port mapping and
        based on that prune the paths to only those local ports
        which are currently mapped to a front panel port
        """
        update_timer = False
        self.mapped_local_ports = self.port_handler.get_all_int_local_ports()
        self._logger.debug(f"spice: mapped_local_ports : {self.mapped_local_ports}")
        for spice_path_obj in self._local_paths:
            port, dummy_name = self.__extract_port_from_path(spice_path_obj.xpath, False)
            if port == 0:
                continue
            if port not in self.mapped_local_ports:
                if spice_path_obj in self._spice_path_objs:
                    self._logger.info("spice: port {} got removed from FP mapping".format(port))
                    self._spice_path_objs.remove(spice_path_obj)
                    update_timer = True
            else:
                if spice_path_obj not in self._spice_path_objs:
                    self._logger.debug("spice: port {} got mapped to FP. Adding to collect".format(port))
                    spice_path_obj.force_reset_skipped_paths()
                    self._spice_path_objs.append(spice_path_obj)
                    update_timer = True

        if update_timer:
            self._log_current_monitored_paths()
            self.__config_path_timers()

    def send_SpiceAmber_data(self, obj_list, spice_ts, is_dump_enabled):
        send_list_data(obj_list, obj_list[0].__class__.__name__)
        self.save_SpiceAmber_data(obj_list, spice_ts, is_dump_enabled)

    def save_SpiceAmber_data(self, obj_list, spice_ts, is_dump_enabled):
        if is_dump_enabled:
            file_name = DefaultPaths.SPICE_DBG_DUMP_DIR+f"spice_dbg_{obj_list[0].__class__.__name__}_{spice_ts}.txt"

            first_obj_list = True
            if os.path.exists(file_name):
                # Remove the preceeding ']' and set marker to add ',\n'
                first_obj_list = False
                with open(file_name, 'r+') as file:
                    file.seek(0, os.SEEK_END)
                    current_position = file.tell()
                    if current_position > 1:
                        file.truncate(current_position - 1)
            else:
                self._logger.debug("writing file {}".format(file_name))

            with open(file_name, "a") as file:
                if not first_obj_list:
                    file.write(",\n")
                else:
                    file.write("[")

                first_obj = True
                for obj in obj_list:
                    if not first_obj:
                        file.write(",\n")
                    else:
                        first_obj = False
                    outstr = json.dumps(obj.to_json())
                    file.write(outstr)
                file.write("]")

    def _handle_spice_path_timer(self, epoch):
        new_timers, self._timers = self._timers.copy(), {}
        obj_lists = [[] for _ in range(self.spice_groups_max)]
        batch_count = [0 for _ in range(self.spice_groups_max)]
        data_exported = [False for _ in range(self.spice_groups_max)]
        index = None
        rc = False
        for path_obj in new_timers:
            if epoch > new_timers[path_obj]:
                self._logger.debug("spice: processing path WL: {} ; epoch : {}".format(path_obj.xpath, str(epoch)))
                path_obj.reset_skipped_paths()
                for path in path_obj.path_set:
                    obj, index = self.__process_spice_path(path_obj, path, epoch)
                    if obj:
                        obj_lists[index].append(obj)
                        batch_count[index] += 1
                        if batch_count[index] == self._DEFAULT_BATCH_COUNT:
                            batch_count[index] = 0
                            self.model_seqno_list[index][self.seqno_batch_idx] += 1
                            self.send_SpiceAmber_data(obj_lists[index], epoch, self.is_spice_dbg_dump_enabled[index])
                            obj_lists[index] = []
                            data_exported[index] = True
                            rc = True

                self.__update_path_timer(path_obj, epoch)
            else:
                self._timers[path_obj] = new_timers[path_obj]

        for index in range(self.spice_groups_max):
            if batch_count[index] != 0 and obj_lists[index]:
                data_exported[index] = True
                self.send_SpiceAmber_data(obj_lists[index], epoch, self.is_spice_dbg_dump_enabled[index])
                obj_lists[index] = []

            if data_exported[index]:
                self.model_seqno_list[index][self.seqno_path_idx] = 0
                self.model_seqno_list[index][self.seqno_batch_idx] = 0
                self.model_seqno_list[index][self.seqno_loop_idx] += 1
                if self.is_spice_dbg_dump_enabled[index] != 0:
                    self.is_spice_dbg_dump_enabled[index] -= 1
        return rc

    def _schedule_update_local_path_list(self):
        if self.port_handler.refresh_port_map() or len(self.mapped_local_ports) == 0:
            self._logger.debug("spice: refreshing the portmap cache")
            self._update_local_path_list()

    def is_spice_dbg_dump_config_updated(self):
        rc = False
        file_path = DefaultPaths.SPICE_CONFIG_FILE_DBG_DUMP

        if os.path.exists(file_path):
            last_modified_time = os.path.getmtime(file_path)
            if last_modified_time != self.spice_dbg_last_modified_time:
                self.spice_dbg_last_modified_time = last_modified_time
                rc = True

        return rc

    def get_spice_dbg_dump_config(self):
        if not self.is_spice_dbg_dump_config_updated():
            return

        file_path = DefaultPaths.SPICE_CONFIG_FILE_DBG_DUMP

        if os.path.exists(file_path):
            with open(file_path) as yaml_file:
                data = yaml.safe_load(yaml_file)
                if data['spice-dbg'] and 'count' in data['spice-dbg']:
                    count = data['spice-dbg']['count']
                    self._logger.debug("spice: dbg capture enabled, count {}".format(count))
                    for i in range(self.spice_groups_max):
                        self.is_spice_dbg_dump_enabled[i] = count

    def create_spice_dbg_dump_dir(self):
        dbg_dir = DefaultPaths.SPICE_DBG_DUMP_DIR
        if not os.path.exists(dbg_dir):
            os.makedirs(dbg_dir)

    def run(self):
        self._schedule_update_local_path_list()
        if not self._is_spice_loaded:
            time.sleep(self._SDK_RESTART_DELAY)
            self._is_spice_loaded = self.__load_spice_module()
            if not self._is_spice_loaded:
                self._logger.error("spice: Not able to load kernel module")
                return
            else:
                self._logger.debug("spice: kernel module loaded")

        self.create_spice_dbg_dump_dir()

        # ensure SDK is up and active before proceeding
        is_active = utils.is_active_systemd_svc('sx_sdk.service')
        if not is_active:
            self._logger.warn("spice: sx_sdk service not alive")
            return
        else:
            self.sx_sdk_pid = utils.get_sx_sdk_pid()

        rc = self.__load_configuration()
        if not rc:
            self._logger.error("spice: Please reach-out to netq-support for any config modifications")
            return

        self._resize_green_pool()

        # yield
        time.sleep(2)
        min_period = self._setup_path_whitelist()

        short_sleep_period = 1
        long_sleep_period = min_period if min_period <= 30 else (min_period-30)

        self.__config_path_timers()

        self.model_seqno_list = [[0] * self.seqno_max_idx for _ in range(self.spice_groups_max)]
        schedule.every(self._CHECK_PORTMAP_INTERVAL).seconds.do(self._schedule_update_local_path_list)

        while True:
            try:
                epoch = time.time()
                self.get_spice_dbg_dump_config()

                current_sx_sdk_pid = utils.get_sx_sdk_pid()
                if current_sx_sdk_pid != self.sx_sdk_pid:
                    self._logger.debug("spice: sdk restart detected; pausing collection for {} seconds".format(
                        self._SDK_RESTART_DELAY))
                    self.sx_sdk_pid = current_sx_sdk_pid
                    time.sleep(self._SDK_RESTART_DELAY)

                # Run the schedule to update local path list every n seconds
                schedule.run_pending()

                # Should check if SDK is active before collecting the counters
                rc = False
                if utils.is_active_systemd_svc('sx_sdk.service'):
                    rc = self._handle_spice_path_timer(epoch)
                if rc:
                    # data was exported, sleep for long as the next schedule is some time away
                    time.sleep(long_sleep_period)
                else:
                    #data was not exported, wake up often to send data
                    time.sleep(short_sleep_period)

            except Exception as e:
                self._logger.error("spice: Process caught Exception {}".format(e))
                break
        return
