#
#  Copyright (C) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#  LicenseRef-NvidiaProprietary
#
#  NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
#  property and proprietary rights in and to this material, related
#  documentation and any modifications thereto. Any use, reproduction,
#  disclosure or distribution of this material and related documentation
#  without an express license agreement from NVIDIA CORPORATION or
#  its affiliates is strictly prohibited.
import json
from netq_agent.parsers.common.render import BaseRender
from netq_lib.common import utils, configuration
from netq_lib.common.enums import ConfKw, DefaultPaths
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate, subprocess_wrapper
from netq_lib.orm.redisdb.models import BoxInfo


class BoxInfoRender(BaseRender):
    __instance = None

    @staticmethod
    def get_instance():
        if BoxInfoRender.__instance is None:
            BoxInfoRender()
        return BoxInfoRender.__instance

    def __init__(self):
        if BoxInfoRender.__instance is None:
            BoxInfoRender.__instance = self
        super(BoxInfoRender, self).__init__()
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        KUBE_CONF = '/etc/kubernetes/admin.conf'
        fullstate = kwargs.get("fullstate")
        box_info_list = []
        try:
            get_nodes_output = subprocess_wrapper("/usr/bin/kubectl --kubeconfig={} get nodes -o wide -o json".format(KUBE_CONF).split())
            # Split the output into lines and extract the 6th column from each line
            node_map = parse_kubectl_output(get_nodes_output)
            cluster_vip_output = get_cluster_vip_output(self.__logger)
            box_info_obj = BoxInfo()
            box_info_obj.nodes = node_map
            box_info_obj.cluster_vip = cluster_vip_output

            if fullstate is True:
                box_info_list.append(box_info_obj)
            else:
                box_info_obj.save()

        except Exception as e:
            self.__logger.error("BoxInfo failed to parse the output. exception: {}".format(e))
            return ServiceStatus.FAIL

        if fullstate:
            save_fullstate(box_info_list, BoxInfo().__class__.__name__)
        return ServiceStatus.OK


def parse_kubectl_output(output):
    # Load the JSON output
    nodes_info = json.loads(output)

    # Extract IP addresses and hostnames
    node_map = {}
    for node in nodes_info["items"]:
        ip = node["status"]["addresses"][0]["address"]
        hostname = node["metadata"]["name"]
        node_map[hostname] = ip
    return node_map


def get_cluster_vip_output(logger):
    cluster_vip = ""
    cluster_vip_file_path = DefaultPaths.CLUSTER_VIP_FILE
    try:
        with open(cluster_vip_file_path, 'r') as file:
            cluster_vip = file.read()
    except Exception as e:
        logger.debug(f"Failed to parse cluster_vip. Error: {e}")
    return cluster_vip


def process_box_info(cmd_output, svc_failed=False, **kwargs):
    box_info_render.parse(cmd_output, svc_failed, **kwargs)


box_info_render = BoxInfoRender.get_instance()
