#
# Copyright (c) 2017-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
import subprocess

from netq_agent.parsers.common.render import BaseRender
from netq_lib.common import configuration, utils
from netq_lib.orm.redisdb.models import Ntp
from netq_lib.common.enums import (
    ConfKw,
    NetqAgentStatus,
    NetqAgentState,
    Platform,
    RunningService,
    CentosRunningService,
    RhelRunningService,
)
from netq_lib.common.utils import (
        is_active_systemd_svc,
        netq_agent_process_status_write,
        get_logger,
        save_fullstate
)

is_msg_logged_once = False
is_ptp_prev_active = False

class NtpRender(BaseRender):
    __instance = None

    @staticmethod
    def get_instance():
        if NtpRender.__instance is None:
            NtpRender()
        return NtpRender.__instance

    def __init__(self):

        if NtpRender.__instance is None:
            NtpRender.__instance = self

        super(NtpRender, self).__init__()
        self._logger = get_logger(ConfKw.NETQ_AGENT, configuration)
        self.__platform = None
        self.ntp_out = None
        self.ntp_cmd = None
        self.use_timedatectl = False
        self.use_chronyc = False
        self.ntp_sync = 'n/a'

    def get_ntp_sync(self):
        return self.ntp_sync

    def set_ntp_sync(self, ntp_sync):
        self.ntp_sync = ntp_sync

    def process_ntpq(self, output, logger, **kwargs):
        global is_msg_logged_once #pylint: disable=global-statement
        global is_ptp_prev_active #pylint: disable=global-statement
        ''' Parse ntpq -p command output
             - Function extracts active server information
             - If activer server is found it returns ntp_sync = 'yes'
        ''' #pylint: disable=pointless-string-statement
        current_server = '-'
        stratum = 16
        ntp = Ntp()
        ntp_sync = 'no'
        is_ptp_active = False
        fullstate = kwargs.get("fullstate")
        ntp_fullstate_arr = []

        if output:
            for line in output.splitlines():
                columns = line.split()
                if len(columns) < 3:
                    continue
                if '***Socket error' in line:
                    continue
                if '*' in columns[0]:
                    current_server = columns[0][1:]
                    if len(columns) >= 8:
                        stratum = columns[-8]
                        ntp_sync = 'yes'
                    else:
                        stratum = 16

        if ntp_sync == 'no':
            if os.path.exists('/bin/systemctl'):
                is_ptp_active = is_active_systemd_svc('ptp4l.service')
                if is_ptp_active:
                    if not is_msg_logged_once:
                        is_msg_logged_once = True
                        msg_string = "NTP not synced... PTP service is active"
                        logger.info(msg_string)
                        netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NTP_STATUS,
                                                        NetqAgentState.NETQ_AGENT_STATE_NA,
                                                        msg_string)
                    if not is_ptp_prev_active:
                        is_ptp_prev_active = True
                else:
                    '''
                    PTP service is not active, handle transition of services
                    say ptp is disabled and ntp was tried to enabled.
                    and NTP sync is 'no', we should set the boolean to log the
                    message to False again.
                    ''' #pylint: disable=pointless-string-statement
                    if is_ptp_prev_active:
                        is_msg_logged_once = False
                        is_ptp_prev_active = False

            if not is_msg_logged_once:
                is_msg_logged_once = True
                logger.info("NTP not synced")
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NTP_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                'NTP not synced, Agent may show up as rotten due to incorrect timestamps')

            if output:
                logger.debug("NTP server details are:")
                for line in output.splitlines():
                    logger.debug("Server: {}".format(line))
        else:
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NTP_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                            'NTP sync success')
        ntp.current_server = current_server
        ntp.stratum = stratum
        ntp.ntp_sync = ntp_sync
        ntp.ntp_app = 'ntpq'
        ntp.ptp_state = is_ptp_active

        if fullstate:
            ntp_fullstate_arr.append(ntp)
            save_fullstate(ntp_fullstate_arr, Ntp().__class__.__name__)
        else:
            ntp.save()

        self.set_ntp_sync(ntp_sync)

    def process_timedatectl(self, output, **kwargs):
        ''' Parse timedatectl status command output
             - checks for 'Network time on: yes' or 'NTP synchronized: yes'
             - checks for 'systemd-timesyncd.service active: yes' (ub1804)
             - If found, returns ntp_sync = 'yes'
             - verified on ub1604 w/ systemd v229
             - verified on ub1804 w/ systemd v237
        '''
        fullstate = kwargs.get("fullstate")
        ntp_fullstate_arr = []
        ntp = Ntp()
        ntp_sync = 'no'

        if output:
            for line in output.splitlines():
                columns = line.split(':')
                # ubuntu and debian datetimectl format differences
                if 'Network time on' in columns[0] or \
                        'NTP synchronized' in columns[0] or \
                        'systemd-timesyncd.service active' in columns[0]:
                    ntp_sync = columns[1].lstrip()
                    break

        # current server and stratum not known
        # when using timedatectl / systemd-timesyncd
        ntp.current_server = '-'
        ntp.stratum = 16
        ntp.ntp_sync = ntp_sync
        ntp.ntp_app = 'timedatectl'
        if fullstate:
            ntp_fullstate_arr.append(ntp)
            save_fullstate(ntp_fullstate_arr, Ntp().__class__.__name__)
        else:
            ntp.save()
        self.set_ntp_sync(ntp_sync)

    def process_chronyc(self, output, **kwargs):
        ''' Parse chronyc sources command output
             - Function extracts active server information
             - If active server is found it returns ntp_sync = 'yes'
             - Eg.
    # chronyc  sources
    210 Number of sources = 4
    MS Name/IP address         Stratum Poll Reach LastRx Last sample
    ===============================================================================
    ^+ up2.com                       2  10   377   733   -815us[ -815us] +/-   95ms
    ^+ ntp1.csl.tjhsst.edu           3   8   377   228   +239us[ +239us] +/-   89ms
    ^? excalibur.prolixium.com       0  10     0     -     +0ns[   +0ns] +/-    0ns
    ^* resolver1.skyfiberintern>     2  10   377   939   +277us[ +404us] +/-   44ms

        '''
        current_server = '-'
        stratum = 16
        fullstate = kwargs.get("fullstate")
        ntp_fullstate_arr = []
        ntp = Ntp()
        ntp_sync = 'no'

        if output:
            for line in output.splitlines():
                columns = line.split()
                if len(columns) < 3:
                    continue
                if '*' in columns[0]:
                    current_server = columns[1]
                    if len(columns) >= 2:
                        stratum = columns[2]
                        ntp_sync = 'yes'
                    else:
                        stratum = 16

        ntp.current_server = current_server
        ntp.stratum = stratum
        ntp.ntp_sync = ntp_sync
        ntp.ntp_app = 'chronyc'
        if fullstate:
            ntp_fullstate_arr.append(ntp)
            save_fullstate(ntp_fullstate_arr, Ntp().__class__.__name__)
        else:
            ntp.save()

        self.set_ntp_sync(ntp_sync)

    def __is_service_running(self, daemon, vrf=None):
        """Check if service is running. Works across systemctl & service
        models.
        """

        if os.path.isfile('/bin/systemctl'):
            if vrf:
                cmdstr = '/bin/systemctl is-active %s@%s' % (daemon, vrf)
            else:
                cmdstr = '/bin/systemctl is-active %s' % daemon
        else:
            cmdstr = '/usr/sbin/service %s status' % daemon

        try:
            utils.subprocess_wrapper(cmdstr.split())
        except (subprocess.CalledProcessError, OSError):
            return False

        return True

    def __set_command(self):
        ntp_out = None
        ntp_cmd = None
        use_timedatectl = False
        use_chronyc = False
        if Platform.CUMULUS in self.__platform:
            ntp_svc = RunningService.NTP
            chrony_svc = RunningService.CHRONY
            if is_active_systemd_svc(ntp_svc+'@mgmt'):
                ntp_cmd = '/usr/bin/vrf task exec mgmt ntpq -p'
            elif is_active_systemd_svc(ntp_svc):
                ntp_cmd = '/usr/bin/ntpq -p'
            elif is_active_systemd_svc(ntp_svc+'@default'):
                ntp_cmd = '/usr/bin/vrf task exec default ntpq -p'
            elif is_active_systemd_svc(chrony_svc):
                if os.path.exists('/usr/bin/chronyc'):
                    ntp_cmd = '/usr/bin/chronyc sources'
                use_chronyc = True
            elif is_active_systemd_svc(chrony_svc+'@mgmt'):
                if os.path.exists('/usr/bin/chronyc'):
                    ntp_cmd = '/usr/bin/vrf task exec mgmt chronyc sources'
                use_chronyc = True
        else:
            # on non-CL platforms, we can support
            # ntpd, chronyd, timedatectl (we check in that order)
            if 'centos' in self.__platform.lower():
                ntp_svc = CentosRunningService.NTP
                chrony_svc = CentosRunningService.CHRONY
            elif 'rhel' in self.__platform.lower():
                ntp_svc = RhelRunningService.NTP
                chrony_svc = RhelRunningService.CHRONY
            elif 'red hat' in self.__platform.lower():
                ntp_svc = RhelRunningService.NTP
                chrony_svc = RhelRunningService.CHRONY
            else:
                ntp_svc = RunningService.NTP
                chrony_svc = RunningService.CHRONY
            if self.__is_service_running(ntp_svc):
                if os.path.exists('/usr/bin/ntpq'):
                    ntp_cmd = '/usr/bin/ntpq -p'
                elif os.path.exists('/usr/sbin/ntpq'):
                    ntp_cmd = '/usr/sbin/ntpq -p'
                elif os.path.exists('/sbin/ntpq'):
                    ntp_cmd = '/sbin/ntpq -p'
            elif (self.__is_service_running(chrony_svc) or
                  self.__is_service_running(chrony_svc, vrf='mgmt')):
                if os.path.exists('/usr/bin/chronyc'):
                    ntp_cmd = '/usr/bin/chronyc sources'
                elif os.path.exists('/bin/chronyc'):
                    ntp_cmd = '/bin/chronyc sources'
                use_chronyc = True
            elif os.path.exists('/bin/timedatectl'):
                ntp_cmd = '/bin/timedatectl status'
                use_timedatectl = True
            elif os.path.exists('/usr/bin/timedatectl'):
                ntp_cmd = '/usr/bin/timedatectl status'
                use_timedatectl = True

        self.ntp_cmd = ntp_cmd
        self.use_timedatectl = use_timedatectl
        self.use_chronyc = use_chronyc

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        self.ntp_out = None
        fullstate = kwargs.get("fullstate")
        self.__platform = kwargs.get('platform', '')

        self.__set_command()

        if self.ntp_cmd:
            try:
                ntp_out = subprocess.check_output(self.ntp_cmd, timeout=45, shell=True)
                self.ntp_out = ntp_out.decode()
            except Exception as e:
                pass

        if self.use_timedatectl:
            self.process_timedatectl(self.ntp_out)
        elif self.use_chronyc:
            self.process_chronyc(self.ntp_out)
        else:
            self.process_ntpq(self.ntp_out, self._logger)

def process_ntp(cmd_output, svc_failed=False, **kwargs):
    return ntp_render.parse(cmd_output, svc_failed, **kwargs)

ntp_render = NtpRender.get_instance()
