#
# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
import re
import csv
import ctypes
import ujson as json
from datetime import datetime
from netq_lib.common.enums import ConfKw
from netq_lib.common import utils, configuration
from netq_agent.parsers.common.render import BaseRender
from netq_lib.common.enums import ServiceStatus
from netq_lib.orm.redisdb.models import BERInfo
from netq_agent.handlers.port_handler import PortHandler
from netq_lib.common.utils import subprocess_wrapper, is_active_systemd_svc, send_list_data
from netq_lib.common.enums import Platform
from netq_agent.handlers.sdk_utils import SX_CHIP_TYPE_SPECTRUM, SX_CHIP_TYPE_SPECTRUM2, SdkUtils

# Regex
DEV_REGX = r'.*_pciconf0\b$'
PORT_REGEX = r'swp(\d+)(?:s(\d*))?'
LOCAL_PORT_REGEX = r'\((\d+)\)'

DEVICE_DIR = '/dev/mst'
SYSTEMCTL = '/bin/systemctl'
SDK_SERVICE = 'sx_sdk.service'
MLX_CMD = '/usr/lib/cumulus/mlxcmd --json ports info'
MLX_LINK = '/usr/bin/mlxlink'
AMBER_COLLECT = '--amber_collect'
BER_COLLECT = '--ber_collect'
CMD = '{} -d {}/{} -p {} -c -j {} /tmp/{}.csv'
CSV_DIR = '/tmp/{}.csv'

# Payload batch size to be sent to GRPC
BATCH_SIZE = 150

# Default values
NOT_AVAILABLE = 'N/A'
PORT_STATUS_DOWN = 'down'
PORT_STATUS_UP = 'Up'
MINUS_ONE = -1
DOUBLE_ZERO = 0.0
MINUS_ONE_STR = '-1'
EMPTY_STR = ''
UTF_ENCODING = 'utf-8'
M_CHAR = 'm'
STATUS_FIELD = 'status'
FEC_FIELD = 'fec'
UNKNOWN_VALUE = 'unknown'
PLATFORM_STR = 'platform'
RESULT_STR = 'result'
OUTPUT_STR = 'output'
VALUES_STR = 'values'
UNSUPPORTED = 'unsupported'
ZERO = '0'

# Fields to extract
EFFECTIVE_BER = 'effective_ber'
DEVICE_ID = 'device_id'
SYMBOL_ERRORS = 'symbol_errors'
DEVICE_FW_VERSION = 'device_fw_version'
MODULE_TEMP = 'module_temp'
ETHERNET_PROTOCOL_ACTIVE = 'ethernet_protocol_active'
PORT_NUMBER = 'port_number'
FC_ZERO_HIST = 'fc_zero_hist'
PHY_MANAGER_STATE = 'phy_manager_state'
CABLE_PN = 'cable_pn'
SPEED = 'speed_[gb/s]'
CABLE_SN = 'cable_sn'
ACTIVE_FEC = 'active_fec'
CABLE_TYPE = 'cable_type'
TIME_SINCE_LAST_CLEAR = 'time_since_last_clear_[min]'
CABLE_LENGTH = 'cable_length'
RAW_BER = 'raw_ber'
ADVANCED_STATUS_OPCODE = 'advanced_status_opcode'

IFNAME = 'ifname'

# Alternative Fields to extract
EFFECTIVE_BER_ALT = 'effective ber'
DEVICE_ID_ALT = 'device id'
DEVICE_FW_VERSION_ALT = 'fw version'
MODULE_TEMP_ALT = 'module start temp'
ETHERNET_PROTOCOL_ACTIVE_ALT = 'protocol'
CABLE_PN_ALT = 'cable pn'
SPEED_ALT = 'speed [gb/s]'
CABLE_SN_ALT = 'cable sn'
ACTIVE_FEC_ALT = 'active fec'
CABLE_TYPE_ALT = 'media'
TIME_SINCE_LAST_CLEAR_ALT = 'test time [min]'
CABLE_LENGTH_ALT = 'length [m]'
RAW_BER_ALT = 'total raw ber'

# Unsupported ASICS SP1/SP1a1 and SP2
UNSUPPORTED_ASICS = (SX_CHIP_TYPE_SPECTRUM, SX_CHIP_TYPE_SPECTRUM2)

def is_valid_device(dev):
    return bool(re.match(DEV_REGX, dev))


def delete_file_if_exist(file_path):
    try:
        os.remove(file_path)
    except OSError:
        pass


def to_double(arg, default=DOUBLE_ZERO):
    try:
        updated_arg = DOUBLE_ZERO if arg is None else arg
        casted_float = ctypes.c_double(float(updated_arg)).value if updated_arg.strip() else default
        return casted_float
    except ValueError:
        return DOUBLE_ZERO


def to_int(arg, default=MINUS_ONE_STR):
    try:
        updated_arg = MINUS_ONE if arg is None else arg
        casted_Int = ctypes.c_int(int(float(updated_arg))).value if updated_arg.strip() else default
        return casted_Int
    except ValueError:
        return MINUS_ONE


def extract_local_port(port_number):
    match = re.search(LOCAL_PORT_REGEX, port_number)
    if match:
        return match.group(1)
    else:
        return port_number


class DataBuffer:
    def __init__(self, threshold_size):
        self.threshold_size = threshold_size
        self.buffer = []
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)

    def add_data(self, data):
        self.buffer.append(data)
        if len(self.buffer) >= self.threshold_size:
            self.flush_buffer()

    def flush_buffer(self):
        if self.buffer:
            send_list_data(self.buffer, BERInfo().__class__.__name__)
            self.buffer = []


class BERRender(BaseRender):
    __logger = None

    def get_logger(self):
        if BERRender.__logger is None:
            __logger = self.__logger
        return BERRender.__logger

    def __init__(self):
        super(BERRender, self).__init__()
        self.model = BERInfo
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)
        self.buffer = None
        self.sdk_utils = SdkUtils(self.__logger)
        self.chip_type = None
        self.port_handler = PortHandler.get_instance(self.__logger, None)
        BERRender.__logger = self.__logger

    def get_up_down_ports(self, ports):
        all_ports = [value[IFNAME] for key, value in ports]
        up_ports = [value[IFNAME] for _, value in ports if
                    value[STATUS_FIELD] == PORT_STATUS_UP and value[STATUS_FIELD] != UNKNOWN_VALUE]
        down_ports = [port for port in all_ports if port not in up_ports]
        return (
        list(filter(lambda x: x[1] != MINUS_ONE_STR, [self.get_mlx_link_port(up_port) for up_port in up_ports])),
        list(down_ports))

    def get_mlx_link_port(self, up_port):
        match = re.match(PORT_REGEX, up_port)
        try:
            if match:
                port_number = match.group(1)
                split = int(match.group(2)) + 1 if match.group(2) else None
                if split is None or split == 1:
                    return [up_port, f'{port_number}']
                else:
                    return [up_port, f'{port_number}/{split}']
            else:
                return [up_port, MINUS_ONE_STR]
        except Exception as e:
            self.__logger.warn('Unable to generate mlx cmd port {} got exception {}'.format(up_port, e))
            return [up_port, MINUS_ONE_STR]

    def get_device(self):
        try:
            dirs = os.listdir(DEVICE_DIR)
            port_path_list = [item for item in dirs if is_valid_device(item)]
            return port_path_list[0]
        except Exception as e:
            self.__logger.warn('Unable to get device for BER, got exception {}'.format(e))
            return ServiceStatus.FAIL

    def get_ports(self):
        if os.path.exists(SYSTEMCTL):
            is_active = is_active_systemd_svc(SDK_SERVICE)
            if not is_active:
                self.__logger.warn(
                    'Unable to get ports for BER, {} is inactive deleting old counters to avoid ambiguity'.format(
                        SDK_SERVICE))
                return ServiceStatus.FAIL
            else:
                try:
                    mlx_port_info = subprocess_wrapper(MLX_CMD.split())
                    return list(json.loads(mlx_port_info).items())
                except Exception as e:
                    self.__logger.warn('Unable to get ports for BER, got exception {}'.format(e))
                    return ServiceStatus.FAIL

    def execute_ber_collect(self, cmd):
        try:
            subprocess_wrapper(cmd.split())
        except Exception as e:
            self.__logger.warn('Unable to execute command for both {} and {}, got exception {}'.format(AMBER_COLLECT, BER_COLLECT, e))
            return ServiceStatus.FAIL

    def generate_ber_info(self, device, up_ports):
        for port in up_ports:
            try:
                cmd = CMD.format(MLX_LINK, DEVICE_DIR, device, port[1], AMBER_COLLECT, port[0])
                delete_file_if_exist(CSV_DIR.format(port[0]))
                subprocess_wrapper(cmd.split())
            except Exception as e:
                cmd2 = CMD.format(MLX_LINK, DEVICE_DIR, device, port[1], BER_COLLECT, port[0])
                self.execute_ber_collect(cmd2)
            finally:
                self.prepare_ber_response(port)

    def prepare_ber_response(self, port):
        try:
            return self.extract_csv_info(CSV_DIR.format(port[0]), port)
        except Exception as e:
            self.__logger.warn('Unable to prepare BER response got exception {}'.format(e))
            return ServiceStatus.FAIL

    def generate_ber_payload(self, effective_ber, device_id, symbol_errors, device_fw_version, module_temp,
                             ethernet_protocol_active, ifname, port_number, local_port, fc_zero_hist,
                             phy_manager_state, cable_pn, speed, cable_sn, active_fec, cable_type,
                             time_since_last_clear,
                             cable_length, raw_ber, advanced_status_opcode, port_status):

        ber_info = BERInfo()
        ber_info.effective_ber = effective_ber
        ber_info.device_id = device_id
        ber_info.symbol_errors = symbol_errors
        ber_info.device_fw_version = device_fw_version
        ber_info.module_temp = module_temp
        ber_info.ethernet_protocol_active = ethernet_protocol_active
        ber_info.ifname = ifname
        ber_info.port_number = port_number
        ber_info.local_port = local_port
        ber_info.fc_zero_hist = fc_zero_hist
        ber_info.phy_manager_state = phy_manager_state
        ber_info.cable_pn = cable_pn
        ber_info.speed = speed
        ber_info.cable_sn = cable_sn
        ber_info.active_fec = active_fec
        ber_info.cable_type = cable_type
        ber_info.time_since_last_clear = time_since_last_clear
        ber_info.cable_length = cable_length
        ber_info.raw_ber = raw_ber
        ber_info.advanced_status_opcode = advanced_status_opcode
        ber_info.port_status = port_status.lower()

        self.buffer.add_data(ber_info)

    def generate_down_port_info(self, down_ports):
        for down_port in down_ports:
            self.generate_ber_payload(
                NOT_AVAILABLE, NOT_AVAILABLE, MINUS_ONE, NOT_AVAILABLE, MINUS_ONE, NOT_AVAILABLE,
                down_port, down_port, MINUS_ONE, MINUS_ONE, NOT_AVAILABLE, NOT_AVAILABLE, MINUS_ONE, NOT_AVAILABLE,
                NOT_AVAILABLE, NOT_AVAILABLE, DOUBLE_ZERO, MINUS_ONE, NOT_AVAILABLE, NOT_AVAILABLE, PORT_STATUS_DOWN
            )

        return ServiceStatus.OK

    def extract_csv_info(self, csv_file_path, port):
        if not os.path.exists(csv_file_path):
            self.__logger.warn('Unable to locate csv file to read {}'.format(csv_file_path))
            return ServiceStatus.FAIL

        with open(csv_file_path, 'r', newline=EMPTY_STR, encoding=UTF_ENCODING) as csv_file:
            reader = csv.DictReader((row.replace('\0', EMPTY_STR).replace('\x00', EMPTY_STR) for row in csv_file),
                                    delimiter=',')
            reader.fieldnames = [name.lower() for name in reader.fieldnames]
            latest_row = None

            for row in reader:
                latest_row = row

            self.generate_ber_payload(
                latest_row.get(EFFECTIVE_BER, latest_row.get(EFFECTIVE_BER_ALT, UNSUPPORTED)),
                latest_row.get(DEVICE_ID, latest_row.get(DEVICE_ID_ALT, UNSUPPORTED)),
                to_int(latest_row.get(SYMBOL_ERRORS, MINUS_ONE_STR)),
                latest_row.get(DEVICE_FW_VERSION, latest_row.get(DEVICE_FW_VERSION_ALT, UNSUPPORTED)),
                to_int(latest_row.get(MODULE_TEMP, latest_row.get(MODULE_TEMP_ALT, MINUS_ONE_STR))),
                latest_row.get(ETHERNET_PROTOCOL_ACTIVE, latest_row.get(ETHERNET_PROTOCOL_ACTIVE_ALT, UNSUPPORTED)),
                port[0],
                port[1],
                to_int(extract_local_port(latest_row.get(PORT_NUMBER, MINUS_ONE_STR))),
                to_int(latest_row.get(FC_ZERO_HIST, MINUS_ONE_STR)),
                latest_row.get(PHY_MANAGER_STATE, UNSUPPORTED),
                latest_row.get(CABLE_PN, latest_row.get(CABLE_PN_ALT, UNSUPPORTED)),
                to_int(latest_row.get(SPEED, latest_row.get(SPEED_ALT, ZERO))),
                latest_row.get(CABLE_SN, latest_row.get(CABLE_SN_ALT, UNSUPPORTED)),
                latest_row.get(ACTIVE_FEC, latest_row.get(ACTIVE_FEC_ALT, UNSUPPORTED)),
                latest_row.get(CABLE_TYPE, latest_row.get(CABLE_TYPE_ALT, UNSUPPORTED)),
                to_double(latest_row.get(TIME_SINCE_LAST_CLEAR, latest_row.get(TIME_SINCE_LAST_CLEAR_ALT, str(DOUBLE_ZERO)))),
                to_int(latest_row.get(CABLE_LENGTH, latest_row.get(CABLE_LENGTH_ALT, MINUS_ONE_STR)).replace(M_CHAR, EMPTY_STR)),
                latest_row.get(RAW_BER, latest_row.get(RAW_BER_ALT, UNSUPPORTED)),
                latest_row.get(ADVANCED_STATUS_OPCODE, UNSUPPORTED),
                PORT_STATUS_UP)

        return ServiceStatus.OK

    def initialize_chip_type(self, platform):
        if self.chip_type is None:
            self.chip_type = self.sdk_utils.get_chip_type_v3(platform)

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        datetimeNOW = datetime.now()
        self.__logger.debug('Going to collect BER info at {}'.format(datetimeNOW))
        device = self.get_device()
        ports = self.get_ports()

        if device is not ServiceStatus.FAIL and ports is not ServiceStatus.FAIL:
            try:
                up_ports, down_ports = self.get_up_down_ports(ports)
                self.buffer = DataBuffer(threshold_size=BATCH_SIZE)
            except Exception as e:
                self.__logger.warn('Unable to get up Ports got exception {}'.format(e))
                return ServiceStatus.FAIL

            self.__logger.debug('Up ports are {} and Down ports are {} at {}'.format(up_ports, down_ports, datetimeNOW))
            self.generate_ber_info(device, up_ports)
            self.generate_down_port_info(down_ports)
            self.buffer.flush_buffer()
            return ServiceStatus.OK
        else:
            return ServiceStatus.FAIL


def process_ber_info(cmd_output, svc_failed=False, **kwargs):
    platform = kwargs.get(PLATFORM_STR, EMPTY_STR)
    ber_render = BERRender()
    ber_render.initialize_chip_type(platform)

    if Platform.MELLANOX in platform and ber_render.chip_type and ber_render.chip_type not in UNSUPPORTED_ASICS:
        return ber_render.parse(cmd_output, svc_failed, **kwargs)
    else:
        ber_render.get_logger().debug('Skipping collecting BER info on Unsupported device {} with chip type {}'.format(platform, ber_render.chip_type))
        return ServiceStatus.OK
