#
# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#
import eventlet
from greenlet import greenlet
from swsscommon import swsscommon
from swsscommon.swsscommon import DBConnector, ConfigDBConnector


class NvosDBRender(object):
    MODEL = None  # Define in derived class

    def __init__(self):
        self._logger = None

    def process(self, logger):
        self._logger = logger
        db_sources = self.db_sources()
        data = {}
        for (db_name, table_name) in db_sources:
            self._logger.debug("Fetch data from table {}:{}".format(db_name, table_name))
            table = self.get_table(db_name, table_name)
            if data.get(db_name):
                data[db_name].update(table)
            else:
                data.update({db_name: table})
        # process retrieved data
        self.process_fullstate(data)

    def get_table(self, db_name, table_name):
        """Read an entire table from DB.
        Args:
            table_name: Table name.
        Returns:
            Table data in a dictionary form of
            { 'row_key': {'column_key': value, ...}, ...}
            or { ('l1_key', 'l2_key', ...): {'column_key': value, ...}, ...} for a multi-key table.
            Empty dictionary if table does not exist.
        """
        data = {}
        db = swsscommon.DBConnector(db_name, 0)
        table = swsscommon.Table(db, table_name)

        if table.get("")[0]:
            data[table_name] = dict(table.get("")[1])
        else:
            for key in table.getKeys():
                temp = dict(table.get(key)[1])
                new_dict = {key: temp}
                if data.get(table_name):
                    data[table_name].update(new_dict)
                else:
                    data.update({table_name: new_dict})

        return data

    def handle_periodic_timeout(self):
        raise NotImplementedError

    def db_sources(self):
        raise NotImplementedError

    def process_fullstate(self, data):
        raise NotImplementedError

    def data_get(self, data, db_name, table_name):
        return data.get(db_name, {}).get(table_name, {})

    def default_callback(self):
        pass


class NvosPerDBClient(ConfigDBConnector):
    def __init__(self, logger, pool):
        super(NvosPerDBClient, self).__init__(use_unix_socket_path=True)
        self._logger = logger
        self._thread_pool = pool
        self._listener_thread = None

    def listen(self):
        self._logger.info("Start listening to DB {}".format(self.db_name))
        # Start listen Redis keyspace event. Pass a callback function to `init` to handle initial table data.
        pubsub = self.get_redis_client(self.db_name).pubsub()
        pubsub.psubscribe(
            "__keyspace@{}__:*".format(self.get_dbid(self.db_name)))

        while True:
            eventlet.sleep()
            item = pubsub.get_message(timeout=1, interrupt_on_signal=True)
            if 'type' not in item:
                # When timeout or interrupted, item will not contain 'type'
                continue
            if item['type'] == 'pmessage':
                key = item['channel'].split(':', 1)[1]
                try:
                    if self.TABLE_NAME_SEPARATOR in key:
                        (table, row) = key.split(self.TABLE_NAME_SEPARATOR, 1)
                    else:
                        table = key
                        row = table
                    if table in self.handlers:
                        client = self.get_redis_client(self.db_name)
                        if item['data'] == 'del':
                            data = 'del'
                        else:
                            data = self.raw_to_typed(client.hgetall(key))
                        # call callback
                        self._logger.debug("Call to callback table {} row {} data {}".format(table, row, data))
                        self._ConfigDBConnector__fire(table, row, data)
                except Exception as ex:
                    self._logger.error("ERROR {}".format(ex))  # Ignore non table-formatted redis entries

    def is_listener_running(self):
        return self._listener_thread is not None

    def __stop_monitor_checker(self, green_thread):
        try:
            green_thread.wait()
        except greenlet.GreenletExit:
            pass
        except Exception as ex:
            self._logger.exception('Monitor thread terminated unexpectedly. Error: %s', ex)
        finally:
            if green_thread == self._listener_thread:
                self._logger.info('Listener thread ended for {}'.format(self.db_name))
                self._listener_thread = None

    def start_listener(self):
        if not self._listener_thread:
            self._listener_thread = self._thread_pool.spawn(self.listen)
            self._listener_thread.link(self.__stop_monitor_checker)

    def stop_listener(self):
        self._logger.info("Stop listener thread {}".format(self.db_name))
        if self._listener_thread:
            self._listener_thread.kill()


class NvosDBSubscribeManager(object):
    def __init__(self, pool, logger):
        self.per_db_client = dict()
        self._logger = logger
        self._thread_pool = pool
        self._registered_models = list()

    def is_model_registered(self, model):
        return model in self._registered_models

    def register_callback(self, render):
        # Check if Model is already registered
        if self.is_model_registered(render.MODEL.__name__):
            return

        # Subscribe for DB sources
        db_sources = render.db_sources()
        for (db_name, table), handler in db_sources.items():
            if db_name not in self.per_db_client:
                self.per_db_client[db_name] = NvosPerDBClient(self._logger, self._thread_pool)
                self.per_db_client[db_name].db_connect(db_name)
            self.per_db_client[db_name].subscribe(table, handler)
            if not self.per_db_client[db_name].is_listener_running():
                self.per_db_client[db_name].start_listener()

        # Mark the model as registered
        self._registered_models.append(render.MODEL.__name__)
        self._logger.info('Registered Models {}'.format(self._registered_models))

    def exit_gracefully(self):
        for _, db_client in self.per_db_client.items():
            db_client.stop_listener()
